(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.Interactor = {}));
})(this, (function (exports) { 'use strict';

  const {
    now
  } = Date;
  const {
    random,
    round
  } = Math;
  const {
    assign,
    create,
    defineProperty,
    defineProperties,
    entries,
    freeze,
    getOwnPropertyDescriptor,
    getOwnPropertyDescriptors,
    getPrototypeOf,
    hasOwn
  } = Object;
  function named(name, fn) {
    return defineProperty(fn, 'name', {
      value: name
    });
  }
  function map(obj, fn) {
    return !obj ? {} : entries(obj).reduce((result, _ref) => {
      let [k, v] = _ref;
      let value = fn(v, k);
      return value != null ? assign(result, {
        [k]: value
      }) : result;
    }, create(null));
  }
  function mapPropertyDescriptors(obj, fn) {
    let descr = getOwnPropertyDescriptors(obj);
    let proto = getPrototypeOf(obj);
    while (proto && proto !== Object.prototype) {
      descr = assign(getOwnPropertyDescriptors(proto), descr);
      proto = getPrototypeOf(proto);
    }
    return map(descr, fn);
  }

  const sym = Symbol('interactor.meta');
  const m = {
    get(inst, key) {
      let meta = inst?.[sym] || create(null);
      return key ? meta[key] : meta;
    },
    set(inst, key, val) {
      if (typeof key === 'string') return m.set(inst, {
        [key]: val
      });
      return defineProperty(inst, sym, {
        enumerable: false,
        configurable: true,
        value: freeze(assign(create(null), {
          id: round(random() * now())
        }, m.get(inst), map(key, (next, key) => {
          let prev = m.get(inst, key);
          switch (key) {
            case 'queue':
              return (prev || []).concat(next);
            case 'keyboard':
            case 'assertions':
            case 'children':
              return assign({}, prev, next);
            default:
              return next;
          }
        })))
      });
    },
    top(inst, topmost) {
      let {
        top,
        parent
      } = m.get(inst);
      return !parent || !topmost && top ? inst : m.top(parent);
    },
    new(inst, key, val) {
      if (key === 'parent') return m.new(inst, {
        parent: val,
        top: false
      });else return m.set(m.set(new inst.constructor(), m.get(inst)), key, val);
    },
    eq(a, b) {
      return m.get(a, 'id') === m.get(b, 'id');
    }
  };

  function InteractorError(message) {
    if (!(this instanceof InteractorError)) return new InteractorError(message);
    Error.prototype.constructor.call(this, message);
    Error.captureStackTrace?.(this, InteractorError);
    defineProperties(this, {
      raw: {
        value: message
      },
      message: {
        configurable: true,
        get: () => this.ctx && this.format().message || this.raw
      }
    });
  }
  defineProperties(InteractorError, {
    prototype: {
      value: create(Error.prototype)
    }
  });
  defineProperties(InteractorError.prototype, {
    constructor: {
      value: InteractorError
    },
    name: {
      get() {
        return this.constructor.name;
      }
    },
    bind: {
      value(ctx, expected) {
        return this.ctx ? this : defineProperties(this, {
          ctx: {
            value: ctx
          },
          expected: {
            value: expected
          }
        });
      }
    },
    format: {
      value() {
        let message = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : '%{e}';
        return defineProperties(this, {
          message: {
            value: format(message.replace('%{e}', this.raw), this.ctx, this.expected)
          }
        });
      }
    }
  });
  const regexDF = /%\{((?:%\{.*?}|.)*?)}/gs;
  const regexNS = /^(true|false|undefined|null|[\d.,]+|\[.*\]|\{.*})$/;
  const regexES = /\s{2,}/g;
  function format(message, inst, expected) {
    return message.replace(regexDF, (_, f) => {
      let arg = f.substr(1).trim();
      if (regexDF.test(arg)) arg = format(arg, inst, expected);
      switch (f[0]) {
        case '@':
          return arg ? `${arg} within ${inst}` : inst.toString();
        case '-':
          return arg.split('|')[expected ? 0 : 1] || '';
        case '"':
          return !regexNS.test(arg) ? `"${arg}"` : arg;
        default:
          return '';
      }
    }).replace(regexES, ' ').trim();
  }

  function assertion(get) {
    let matcher = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : get;
    get = matcher === get ? null : get;
    return function (expected) {
      for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
        args[_key - 1] = arguments[_key];
      }
      if (get) {
        let value = get.apply(this, args.slice(0, get.length));
        args = [value].concat(args);
      }
      let {
        message,
        result
      } = matcher.apply(this, args);
      if (result !== expected) throw InteractorError(message);
    };
  }
  function createAssert(name, fn) {
    return named(name, assertion(fn, function (actual) {
      for (var _len2 = arguments.length, args = new Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
        args[_key2 - 1] = arguments[_key2];
      }
      let what = fn?.length ? args[0] : name;
      let message = `%{@} ${what} is %{"${actual}}`;
      let expected = args[fn?.length || 0];
      let result = !!actual;
      if (expected != null) {
        message += ` but expected %{- %{"${expected}}|it not to be}`;
        if (typeof actual === 'string' && expected instanceof RegExp) {
          result = expected.test(actual);
        } else if (typeof expected === 'function') {
          result = expected.call(this, actual);
          result = typeof result === 'undefined' || result;
        } else {
          result = actual === expected;
        }
      } else if (typeof actual === 'boolean') {
        message = `%{@} is %{- not} ${what}`;
      }
      return {
        message,
        result
      };
    }));
  }
  function ctxError() {
    throw InteractorError('interactor methods should not be called within assertions');
  }
  function mapAssertMethods(assert, expected, ctx) {
    let negated = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : !expected;
    return (fn, name) => ({
      value: function () {
        for (var _len3 = arguments.length, args = new Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
          args[_key3] = arguments[_key3];
        }
        return assert(named(`assert${negated ? '.not' : ''}.${name}`, function () {
          let c = ctx || context(this, expected);
          for (var _len4 = arguments.length, a = new Array(_len4), _key4 = 0; _key4 < _len4; _key4++) {
            a[_key4] = arguments[_key4];
          }
          c.assert(fn.bind(c, expected, ...args, ...a));
        }));
      }
    });
  }
  function mapAssertChildren(assert, inst, negated) {
    return (child, name) => ({
      [typeof child === 'function' ? 'value' : 'get']: named(`assert${negated ? '.not' : ''}.${name}`, function () {
        for (var _len5 = arguments.length, args = new Array(_len5), _key5 = 0; _key5 < _len5; _key5++) {
          args[_key5] = arguments[_key5];
        }
        return assert(m.new(typeof child === 'function' ? child.apply(m.new(inst, 'top', true), args) : child, 'parent', inst));
      })
    });
  }
  function context(i, expected, negated) {
    let proto = i.constructor.prototype;
    let {
      assertions,
      children
    } = m.get(proto.assert);
    let ctx = create(i, mapPropertyDescriptors(proto, (_ref, key) => {
      let {
        get,
        value
      } = _ref;
      if (['constructor', '$', '$$', 'assert'].includes(key)) return;
      if (['timeout', 'exec', 'catch', 'then'].includes(key)) return {
        value: ctxError
      };
      return {
        [get ? 'get' : 'value']() {
          for (var _len6 = arguments.length, args = new Array(_len6), _key6 = 0; _key6 < _len6; _key6++) {
            args[_key6] = arguments[_key6];
          }
          let ret = (get || value).apply(this, args);
          return m.get(ret, 'queue') ? context(ret, expected) : ret;
        }
      };
    }));
    defineProperties(ctx, {
      assert: {
        value(f) {
          if (m.get(f, 'queue')) return context(f, expected, negated).assert;
          try {
            f.apply(this);
          } catch (err) {
            if (err.name === 'InteractorError') err.bind(this, expected);
            throw err;
          }
          return this;
        }
      }
    });
    defineProperties(ctx.assert, assign(map(assertions, mapAssertMethods(f => f(), expected, ctx, negated)), map(children, mapAssertChildren(ctx.assert, i, negated))));
    if (!negated) {
      defineProperty(ctx.assert, 'not', {
        get: () => context(i, !expected, true).assert
      });
    }
    return ctx;
  }
  function InteractorAssert(i) {
    let expected = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : true;
    let passert = i.constructor.prototype.assert;
    let {
      assertions,
      children
    } = m.get(passert);
    let assert = function assert(f) {
      if (m.get(f, 'queue')) return InteractorAssert(f, expected);
      return passert.call(i, defineProperties(function () {
        let message;
        try {
          let res = f.apply(context(this, expected), arguments);
          if (typeof res === 'object' && 'message' in res) message = res.message;
          if (typeof res === 'object' && 'result' in res) res = !!res.result;
          if (expected === !!(res ?? true)) return;
        } catch (err) {
          if (err.name === 'InteractorError') err.bind(this, expected);
          if (expected) throw err;
          return;
        }
        throw InteractorError(message ?? ['expected assertion to', expected ? 'pass' : 'fail', 'but it', expected ? 'failed' : 'passed'].join(' '));
      }, {
        name: {
          value: f.name
        },
        length: {
          value: f.length
        }
      }));
    };
    defineProperties(assert, assign(map(assertions, mapAssertMethods(passert.bind(i), expected)), map(children, mapAssertChildren(assert, i, !expected))));
    if (expected) {
      defineProperties(assert, {
        not: {
          get: () => InteractorAssert(i, false)
        },
        remains: {
          value: function remains() {
            let ms = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 500;
            let q = m.get(i, 'queue');
            if (q[q.length - 1]?.type !== 'assert') throw InteractorError('no previous assertion to persist');
            return m.new(i, 'queue', {
              type: 'assert',
              remains: ms
            });
          }
        }
      });
    }
    return assert;
  }

  function when(assertion) {
    let options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    if (Number.isInteger(options)) options = {
      timeout: options
    };
    let {
      timeout = 2000,
      interval = 10,
      remains
    } = options;
    let then = function () {
      return new Promise((resolve, reject) => async function retry(t, r) {
        try {
          let ret = await assertion();
          if (remains && (!r || now() - r <= remains)) setTimeout(retry, interval, t, r || now());else resolve(ret);
        } catch (error) {
          if (r || now() - t <= timeout) setTimeout(retry, interval, t);else reject(error);
        }
      }(now())).then(...arguments);
    };
    return assign(then, {
      catch: f => then().catch(f),
      then
    });
  }

  const RESERVED_PROPERTIES = ['$', '$$', 'assert', 'remains', 'not', 'exec', 'catch', 'then'];
  function reserved(name) {
    if (RESERVED_PROPERTIES.includes(name)) {
      console.warn(`\`${name}\` is a reserved property and will be ignored`);
      return true;
    }
  }
  function isInteractorFn(fn) {
    return typeof fn === 'function' && ('extend' in fn && 'find' in fn && 'assert' in fn.prototype && 'exec' in fn.prototype && 'then' in fn.prototype || isInteractorFn(m.get(fn, 'constructor')));
  }
  function isInteractorDescriptor(prop) {
    let proto = getPrototypeOf(prop);
    return (proto === Object.prototype || proto == null) && ('child' in prop || 'value' in prop || 'get' in prop || 'assert' in prop);
  }
  function wrapInteractorProperty(name, fn, getter) {
    return {
      enumerable: true,
      configurable: true,
      [getter ? 'get' : 'value']: m.set(named(name, function () {
        let top = m.get(this, 'top');
        let result = fn.apply(top ? this : m.new(this, 'top', true), arguments);
        let queue = m.get(result, 'queue');
        if (queue) {
          if (m.eq(this, result)) {
            result = m.new(result, 'top', top);
          } else {
            result = m.new(result, 'parent', this);
          }
          if (!m.get(result, 'top') && queue.length && !getter) {
            while (m.get(result, 'parent')) result = m.get(result, 'parent').exec(result);
          }
        }
        return result;
      }), m.get(fn))
    };
  }
  function defineInteractorProperty(proto, name, descr) {
    let {
      assert,
      child,
      get,
      value
    } = descr;
    if (get || value || typeof child === 'function') {
      defineProperty(proto, name, wrapInteractorProperty(name, get || value || child, !!get));
    } else if (child) {
      defineProperty(proto, name, wrapInteractorProperty(name, () => child, true));
    }
    if (assert !== false && (assert || get || value)) {
      m.set(proto.assert, 'assertions', {
        [name]: assert || createAssert(name, get || value)
      });
    } else if (child) {
      m.set(proto.assert, 'children', {
        [name]: child
      });
    }
    return proto;
  }
  function defineInteractorProperties(proto, properties) {
    let {
      assert,
      ...descriptors
    } = getOwnPropertyDescriptors(properties);
    if (assert) {
      m.set(proto.assert, 'assertions', map(assert.value, (p, k) => reserved(k) ? null : p));
    }
    for (let [key, {
      get,
      value
    }] of entries(descriptors)) {
      if (reserved(key)) continue;
      let property;
      if (get) property = {
        get
      };else if (isInteractorFn(value)) property = {
        child: s => value(s)
      };else if (typeof value === 'function') property = {
        value,
        assert: false
      };else if (m.get(value, 'queue')) property = {
        child: value
      };else if (isInteractorDescriptor(value)) property = value;
      if (property && properties.assert?.[key]) property.assert = false;
      if (property) defineInteractorProperty(proto, key, property);
    }
    return proto;
  }
  function extend() {
    let options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    let properties = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : options;
    if (options === properties) options = null;
    let Parent = this;
    function Extended(selector, props) {
      if (props) return Extended.extend(props)(selector);else if (!(this instanceof Extended)) return new Extended(selector);
      Parent.call(this, selector);
    }
    defineProperties(Extended, assign(getOwnPropertyDescriptors(Parent), {
      prototype: {
        value: create(Parent.prototype, {
          constructor: {
            value: Extended
          },
          assert: {
            value: m.set(function () {
              return Parent.prototype.assert.apply(this, arguments);
            }, m.get(Parent.prototype.assert))
          }
        })
      }
    }));
    assign(Extended, options);
    defineInteractorProperties(Extended.prototype, properties);
    return Extended;
  }

  function dom(inst) {
    return inst.constructor.dom;
  }
  function hasLayoutEngine(inst, subj) {
    let {
      suppressLayoutEngineWarning
    } = m.top(inst).constructor;
    let {
      result,
      debounce
    } = hasLayoutEngine;
    if (result == null) {
      result = hasLayoutEngine.result = !dom(inst).navigator.userAgent.includes('jsdom');
      debounce = hasLayoutEngine.debounce = debounce ?? {};
    }
    if (!suppressLayoutEngineWarning && !result) {
      if (!debounce[subj]) {
        console.warn(`No layout engine detected. ${subj} as a result of CSS cannot be determined. ` + 'You can disable this warning by setting `I.suppressLayoutEngineWarning = true`');
      }
      clearTimeout(debounce[subj]);
      debounce[subj] = setTimeout(() => debounce[subj] = null, 100);
    }
    return result;
  }
  function maybeSelector(selector) {
    for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      args[_key - 1] = arguments[_key];
    }
    let result = selector(...args);
    return typeof result !== 'string' ? [selector, result] : [result];
  }
  function getParentElement(inst, deep) {
    let {
      parent,
      selector,
      nested
    } = m.get(inst);
    let $parent = nested && parent ? parent.$() : dom(inst).document.body;
    let ret = $parent;
    if (typeof selector === 'function') [selector, ret] = maybeSelector(selector, $parent);
    if (selector && typeof selector === 'string' && selector !== '&') ret = $parent.querySelector(selector);
    if (!ret) throw InteractorError('could not find %{@}').bind(inst);
    if (deep && 'contentDocument' in ret) {
      if (!ret.contentDocument) throw InteractorError('%{@} is inaccessible, possibly due to CORS').bind(inst);else ret = ret.contentDocument.body;
    }
    return ret;
  }
  function query(sel, multiple) {
    if (multiple && !sel) {
      throw InteractorError('cannot query for multiple elements without a selector');
    }
    let sq = m.get(sel, 'queue');
    if (sq?.length) {
      throw InteractorError('the provided interactor must not have queued actions');
    }
    let ret = getParentElement(this, !!sel);
    sel = sq ? m.get(sel, 'selector') : sel;
    if (sel) {
      if (typeof sel === 'function') [sel, ret] = maybeSelector(sel, ret, multiple);
      if (typeof sel === 'string') {
        ret = multiple ? ret.querySelectorAll(sel) : ret.querySelector(sel);
      } else if (typeof sel !== 'function') {
        throw InteractorError(`unknown selector: ${sel}`);
      }
    }
    ret = multiple ? Array.from(ret) : ret;
    if (!multiple && !ret) throw InteractorError(`could not find %{@ ${sel}}`).bind(this);
    return ret;
  }
  function dispatch($el, name, _ref) {
    let {
      bubbles = true,
      cancelable = true,
      ...options
    } = _ref;
    return $el.dispatchEvent(assign(new Event(name, {
      bubbles,
      cancelable
    }), options));
  }

  function blur() {
    return this.assert.focused().exec($element => $element.blur());
  }

  function isCheckboxOrRadio($el) {
    if ($el.type !== 'checkbox' && $el.type !== 'radio') throw InteractorError('%{@} is not a checkbox or radio button');
  }
  function check() {
    return this.assert(isCheckboxOrRadio).assert.not.disabled().assert.not.checked().exec($el => $el.click());
  }

  function click() {
    return this.assert.not.disabled().exec($element => $element.click());
  }

  function exec$3($element) {
    let win = $element.ownerDocument.defaultView;
    let {
      document,
      self,
      top,
      frameElement
    } = win;
    if (frameElement && !document.hasFocus() && self !== top) frameElement.focus();
    if (!document.hasFocus()) throw InteractorError('the document is not focusable');
    $element.focus();
  }
  function focus() {
    return this.assert.focusable().exec(exec$3);
  }

  const k = {
    def(key) {
      let def = DEFINITIONS[key];
      if (!def) throw InteractorError(`unknown key \`${key}\``);
      return def;
    },
    press(key, active, next) {
      return m.new(next, 'keyboard', {
        [k.def(key).code]: active
      });
    },
    parse(inst, key) {
      let pressed = m.get(m.top(inst, true), 'keyboard');
      let def = k.def(key);
      let {
        Alt: altKey = def.key === 'Alt' || false,
        Control: ctrlKey = def.key === 'Control' || false,
        Meta: metaKey = def.key === 'Meta' || false,
        Shift: shiftKey = def.key === 'Shift' || false
      } = entries(pressed).reduce((m, _ref) => {
        let [c, v] = _ref;
        return assign(m, {
          [DEFINITIONS[c].key]: v || null
        });
      }, {});
      let event = {
        code: def.code,
        location: def.location || 0,
        repeat: pressed[def.code] || false,
        key: shiftKey && def.shiftKey || def.key,
        keyCode: shiftKey && def.shiftKeyCode || def.keyCode,
        altKey,
        ctrlKey,
        metaKey,
        shiftKey
      };
      let text = !(altKey || ctrlKey || metaKey) && (event.key.length === 1 ? event.key : def.text) || '';
      return {
        text,
        event
      };
    }
  };
  const DEFINITIONS = {
    '0': {
      keyCode: 48,
      key: '0',
      code: 'Digit0'
    },
    '1': {
      keyCode: 49,
      key: '1',
      code: 'Digit1'
    },
    '2': {
      keyCode: 50,
      key: '2',
      code: 'Digit2'
    },
    '3': {
      keyCode: 51,
      key: '3',
      code: 'Digit3'
    },
    '4': {
      keyCode: 52,
      key: '4',
      code: 'Digit4'
    },
    '5': {
      keyCode: 53,
      key: '5',
      code: 'Digit5'
    },
    '6': {
      keyCode: 54,
      key: '6',
      code: 'Digit6'
    },
    '7': {
      keyCode: 55,
      key: '7',
      code: 'Digit7'
    },
    '8': {
      keyCode: 56,
      key: '8',
      code: 'Digit8'
    },
    '9': {
      keyCode: 57,
      key: '9',
      code: 'Digit9'
    },
    'Power': {
      key: 'Power',
      code: 'Power'
    },
    'Eject': {
      key: 'Eject',
      code: 'Eject'
    },
    'Abort': {
      keyCode: 3,
      code: 'Abort',
      key: 'Cancel'
    },
    'Help': {
      keyCode: 6,
      code: 'Help',
      key: 'Help'
    },
    'Backspace': {
      keyCode: 8,
      code: 'Backspace',
      key: 'Backspace'
    },
    'Tab': {
      keyCode: 9,
      code: 'Tab',
      key: 'Tab'
    },
    'Numpad5': {
      keyCode: 12,
      shiftKeyCode: 101,
      key: 'Clear',
      code: 'Numpad5',
      shiftKey: '5',
      location: 3
    },
    'NumpadEnter': {
      keyCode: 13,
      code: 'NumpadEnter',
      key: 'Enter',
      text: '\r',
      location: 3
    },
    'Enter': {
      keyCode: 13,
      code: 'Enter',
      key: 'Enter',
      text: '\r'
    },
    '\r': {
      keyCode: 13,
      code: 'Enter',
      key: 'Enter',
      text: '\r'
    },
    '\n': {
      keyCode: 13,
      code: 'Enter',
      key: 'Enter',
      text: '\r'
    },
    'ShiftLeft': {
      keyCode: 16,
      code: 'ShiftLeft',
      key: 'Shift',
      location: 1
    },
    'ShiftRight': {
      keyCode: 16,
      code: 'ShiftRight',
      key: 'Shift',
      location: 2
    },
    'ControlLeft': {
      keyCode: 17,
      code: 'ControlLeft',
      key: 'Control',
      location: 1
    },
    'ControlRight': {
      keyCode: 17,
      code: 'ControlRight',
      key: 'Control',
      location: 2
    },
    'AltLeft': {
      keyCode: 18,
      code: 'AltLeft',
      key: 'Alt',
      location: 1
    },
    'AltRight': {
      keyCode: 18,
      code: 'AltRight',
      key: 'Alt',
      location: 2
    },
    'Pause': {
      keyCode: 19,
      code: 'Pause',
      key: 'Pause'
    },
    'CapsLock': {
      keyCode: 20,
      code: 'CapsLock',
      key: 'CapsLock'
    },
    'Escape': {
      keyCode: 27,
      code: 'Escape',
      key: 'Escape'
    },
    'Convert': {
      keyCode: 28,
      code: 'Convert',
      key: 'Convert'
    },
    'NonConvert': {
      keyCode: 29,
      code: 'NonConvert',
      key: 'NonConvert'
    },
    'Space': {
      keyCode: 32,
      code: 'Space',
      key: ' '
    },
    'Numpad9': {
      keyCode: 33,
      shiftKeyCode: 105,
      key: 'PageUp',
      code: 'Numpad9',
      shiftKey: '9',
      location: 3
    },
    'PageUp': {
      keyCode: 33,
      code: 'PageUp',
      key: 'PageUp'
    },
    'Numpad3': {
      keyCode: 34,
      shiftKeyCode: 99,
      key: 'PageDown',
      code: 'Numpad3',
      shiftKey: '3',
      location: 3
    },
    'PageDown': {
      keyCode: 34,
      code: 'PageDown',
      key: 'PageDown'
    },
    'End': {
      keyCode: 35,
      code: 'End',
      key: 'End'
    },
    'Numpad1': {
      keyCode: 35,
      shiftKeyCode: 97,
      key: 'End',
      code: 'Numpad1',
      shiftKey: '1',
      location: 3
    },
    'Home': {
      keyCode: 36,
      code: 'Home',
      key: 'Home'
    },
    'Numpad7': {
      keyCode: 36,
      shiftKeyCode: 103,
      key: 'Home',
      code: 'Numpad7',
      shiftKey: '7',
      location: 3
    },
    'ArrowLeft': {
      keyCode: 37,
      code: 'ArrowLeft',
      key: 'ArrowLeft'
    },
    'Numpad4': {
      keyCode: 37,
      shiftKeyCode: 100,
      key: 'ArrowLeft',
      code: 'Numpad4',
      shiftKey: '4',
      location: 3
    },
    'Numpad8': {
      keyCode: 38,
      shiftKeyCode: 104,
      key: 'ArrowUp',
      code: 'Numpad8',
      shiftKey: '8',
      location: 3
    },
    'ArrowUp': {
      keyCode: 38,
      code: 'ArrowUp',
      key: 'ArrowUp'
    },
    'ArrowRight': {
      keyCode: 39,
      code: 'ArrowRight',
      key: 'ArrowRight'
    },
    'Numpad6': {
      keyCode: 39,
      shiftKeyCode: 102,
      key: 'ArrowRight',
      code: 'Numpad6',
      shiftKey: '6',
      location: 3
    },
    'Numpad2': {
      keyCode: 40,
      shiftKeyCode: 98,
      key: 'ArrowDown',
      code: 'Numpad2',
      shiftKey: '2',
      location: 3
    },
    'ArrowDown': {
      keyCode: 40,
      code: 'ArrowDown',
      key: 'ArrowDown'
    },
    'Select': {
      keyCode: 41,
      code: 'Select',
      key: 'Select'
    },
    'Open': {
      keyCode: 43,
      code: 'Open',
      key: 'Execute'
    },
    'PrintScreen': {
      keyCode: 44,
      code: 'PrintScreen',
      key: 'PrintScreen'
    },
    'Insert': {
      keyCode: 45,
      code: 'Insert',
      key: 'Insert'
    },
    'Numpad0': {
      keyCode: 45,
      shiftKeyCode: 96,
      key: 'Insert',
      code: 'Numpad0',
      shiftKey: '0',
      location: 3
    },
    'Delete': {
      keyCode: 46,
      code: 'Delete',
      key: 'Delete'
    },
    'NumpadDecimal': {
      keyCode: 46,
      shiftKeyCode: 110,
      code: 'NumpadDecimal',
      key: '\u0000',
      shiftKey: '.',
      location: 3
    },
    'Digit0': {
      keyCode: 48,
      code: 'Digit0',
      shiftKey: ')',
      key: '0'
    },
    'Digit1': {
      keyCode: 49,
      code: 'Digit1',
      shiftKey: '!',
      key: '1'
    },
    'Digit2': {
      keyCode: 50,
      code: 'Digit2',
      shiftKey: '@',
      key: '2'
    },
    'Digit3': {
      keyCode: 51,
      code: 'Digit3',
      shiftKey: '#',
      key: '3'
    },
    'Digit4': {
      keyCode: 52,
      code: 'Digit4',
      shiftKey: '$',
      key: '4'
    },
    'Digit5': {
      keyCode: 53,
      code: 'Digit5',
      shiftKey: '%',
      key: '5'
    },
    'Digit6': {
      keyCode: 54,
      code: 'Digit6',
      shiftKey: '^',
      key: '6'
    },
    'Digit7': {
      keyCode: 55,
      code: 'Digit7',
      shiftKey: '&',
      key: '7'
    },
    'Digit8': {
      keyCode: 56,
      code: 'Digit8',
      shiftKey: '*',
      key: '8'
    },
    'Digit9': {
      keyCode: 57,
      code: 'Digit9',
      shiftKey: '(',
      key: '9'
    },
    'KeyA': {
      keyCode: 65,
      code: 'KeyA',
      shiftKey: 'A',
      key: 'a'
    },
    'KeyB': {
      keyCode: 66,
      code: 'KeyB',
      shiftKey: 'B',
      key: 'b'
    },
    'KeyC': {
      keyCode: 67,
      code: 'KeyC',
      shiftKey: 'C',
      key: 'c'
    },
    'KeyD': {
      keyCode: 68,
      code: 'KeyD',
      shiftKey: 'D',
      key: 'd'
    },
    'KeyE': {
      keyCode: 69,
      code: 'KeyE',
      shiftKey: 'E',
      key: 'e'
    },
    'KeyF': {
      keyCode: 70,
      code: 'KeyF',
      shiftKey: 'F',
      key: 'f'
    },
    'KeyG': {
      keyCode: 71,
      code: 'KeyG',
      shiftKey: 'G',
      key: 'g'
    },
    'KeyH': {
      keyCode: 72,
      code: 'KeyH',
      shiftKey: 'H',
      key: 'h'
    },
    'KeyI': {
      keyCode: 73,
      code: 'KeyI',
      shiftKey: 'I',
      key: 'i'
    },
    'KeyJ': {
      keyCode: 74,
      code: 'KeyJ',
      shiftKey: 'J',
      key: 'j'
    },
    'KeyK': {
      keyCode: 75,
      code: 'KeyK',
      shiftKey: 'K',
      key: 'k'
    },
    'KeyL': {
      keyCode: 76,
      code: 'KeyL',
      shiftKey: 'L',
      key: 'l'
    },
    'KeyM': {
      keyCode: 77,
      code: 'KeyM',
      shiftKey: 'M',
      key: 'm'
    },
    'KeyN': {
      keyCode: 78,
      code: 'KeyN',
      shiftKey: 'N',
      key: 'n'
    },
    'KeyO': {
      keyCode: 79,
      code: 'KeyO',
      shiftKey: 'O',
      key: 'o'
    },
    'KeyP': {
      keyCode: 80,
      code: 'KeyP',
      shiftKey: 'P',
      key: 'p'
    },
    'KeyQ': {
      keyCode: 81,
      code: 'KeyQ',
      shiftKey: 'Q',
      key: 'q'
    },
    'KeyR': {
      keyCode: 82,
      code: 'KeyR',
      shiftKey: 'R',
      key: 'r'
    },
    'KeyS': {
      keyCode: 83,
      code: 'KeyS',
      shiftKey: 'S',
      key: 's'
    },
    'KeyT': {
      keyCode: 84,
      code: 'KeyT',
      shiftKey: 'T',
      key: 't'
    },
    'KeyU': {
      keyCode: 85,
      code: 'KeyU',
      shiftKey: 'U',
      key: 'u'
    },
    'KeyV': {
      keyCode: 86,
      code: 'KeyV',
      shiftKey: 'V',
      key: 'v'
    },
    'KeyW': {
      keyCode: 87,
      code: 'KeyW',
      shiftKey: 'W',
      key: 'w'
    },
    'KeyX': {
      keyCode: 88,
      code: 'KeyX',
      shiftKey: 'X',
      key: 'x'
    },
    'KeyY': {
      keyCode: 89,
      code: 'KeyY',
      shiftKey: 'Y',
      key: 'y'
    },
    'KeyZ': {
      keyCode: 90,
      code: 'KeyZ',
      shiftKey: 'Z',
      key: 'z'
    },
    'MetaLeft': {
      keyCode: 91,
      code: 'MetaLeft',
      key: 'Meta',
      location: 1
    },
    'MetaRight': {
      keyCode: 92,
      code: 'MetaRight',
      key: 'Meta',
      location: 2
    },
    'ContextMenu': {
      keyCode: 93,
      code: 'ContextMenu',
      key: 'ContextMenu'
    },
    'NumpadMultiply': {
      keyCode: 106,
      code: 'NumpadMultiply',
      key: '*',
      location: 3
    },
    'NumpadAdd': {
      keyCode: 107,
      code: 'NumpadAdd',
      key: '+',
      location: 3
    },
    'NumpadSubtract': {
      keyCode: 109,
      code: 'NumpadSubtract',
      key: '-',
      location: 3
    },
    'NumpadDivide': {
      keyCode: 111,
      code: 'NumpadDivide',
      key: '/',
      location: 3
    },
    'F1': {
      keyCode: 112,
      code: 'F1',
      key: 'F1'
    },
    'F2': {
      keyCode: 113,
      code: 'F2',
      key: 'F2'
    },
    'F3': {
      keyCode: 114,
      code: 'F3',
      key: 'F3'
    },
    'F4': {
      keyCode: 115,
      code: 'F4',
      key: 'F4'
    },
    'F5': {
      keyCode: 116,
      code: 'F5',
      key: 'F5'
    },
    'F6': {
      keyCode: 117,
      code: 'F6',
      key: 'F6'
    },
    'F7': {
      keyCode: 118,
      code: 'F7',
      key: 'F7'
    },
    'F8': {
      keyCode: 119,
      code: 'F8',
      key: 'F8'
    },
    'F9': {
      keyCode: 120,
      code: 'F9',
      key: 'F9'
    },
    'F10': {
      keyCode: 121,
      code: 'F10',
      key: 'F10'
    },
    'F11': {
      keyCode: 122,
      code: 'F11',
      key: 'F11'
    },
    'F12': {
      keyCode: 123,
      code: 'F12',
      key: 'F12'
    },
    'F13': {
      keyCode: 124,
      code: 'F13',
      key: 'F13'
    },
    'F14': {
      keyCode: 125,
      code: 'F14',
      key: 'F14'
    },
    'F15': {
      keyCode: 126,
      code: 'F15',
      key: 'F15'
    },
    'F16': {
      keyCode: 127,
      code: 'F16',
      key: 'F16'
    },
    'F17': {
      keyCode: 128,
      code: 'F17',
      key: 'F17'
    },
    'F18': {
      keyCode: 129,
      code: 'F18',
      key: 'F18'
    },
    'F19': {
      keyCode: 130,
      code: 'F19',
      key: 'F19'
    },
    'F20': {
      keyCode: 131,
      code: 'F20',
      key: 'F20'
    },
    'F21': {
      keyCode: 132,
      code: 'F21',
      key: 'F21'
    },
    'F22': {
      keyCode: 133,
      code: 'F22',
      key: 'F22'
    },
    'F23': {
      keyCode: 134,
      code: 'F23',
      key: 'F23'
    },
    'F24': {
      keyCode: 135,
      code: 'F24',
      key: 'F24'
    },
    'NumLock': {
      keyCode: 144,
      code: 'NumLock',
      key: 'NumLock'
    },
    'ScrollLock': {
      keyCode: 145,
      code: 'ScrollLock',
      key: 'ScrollLock'
    },
    'AudioVolumeMute': {
      keyCode: 173,
      code: 'AudioVolumeMute',
      key: 'AudioVolumeMute'
    },
    'AudioVolumeDown': {
      keyCode: 174,
      code: 'AudioVolumeDown',
      key: 'AudioVolumeDown'
    },
    'AudioVolumeUp': {
      keyCode: 175,
      code: 'AudioVolumeUp',
      key: 'AudioVolumeUp'
    },
    'MediaTrackNext': {
      keyCode: 176,
      code: 'MediaTrackNext',
      key: 'MediaTrackNext'
    },
    'MediaTrackPrevious': {
      keyCode: 177,
      code: 'MediaTrackPrevious',
      key: 'MediaTrackPrevious'
    },
    'MediaStop': {
      keyCode: 178,
      code: 'MediaStop',
      key: 'MediaStop'
    },
    'MediaPlayPause': {
      keyCode: 179,
      code: 'MediaPlayPause',
      key: 'MediaPlayPause'
    },
    'Semicolon': {
      keyCode: 186,
      code: 'Semicolon',
      shiftKey: ':',
      key: ';'
    },
    'Equal': {
      keyCode: 187,
      code: 'Equal',
      shiftKey: '+',
      key: '='
    },
    'NumpadEqual': {
      keyCode: 187,
      code: 'NumpadEqual',
      key: '=',
      location: 3
    },
    'Comma': {
      keyCode: 188,
      code: 'Comma',
      shiftKey: '<',
      key: ','
    },
    'Minus': {
      keyCode: 189,
      code: 'Minus',
      shiftKey: '_',
      key: '-'
    },
    'Period': {
      keyCode: 190,
      code: 'Period',
      shiftKey: '>',
      key: '.'
    },
    'Slash': {
      keyCode: 191,
      code: 'Slash',
      shiftKey: '?',
      key: '/'
    },
    'Backquote': {
      keyCode: 192,
      code: 'Backquote',
      shiftKey: '~',
      key: '`'
    },
    'BracketLeft': {
      keyCode: 219,
      code: 'BracketLeft',
      shiftKey: '{ ',
      key: '['
    },
    'Backslash': {
      keyCode: 220,
      code: 'Backslash',
      shiftKey: '|',
      key: '\\'
    },
    'BracketRight': {
      keyCode: 221,
      code: 'BracketRight',
      shiftKey: ' }',
      key: ']'
    },
    'Quote': {
      keyCode: 222,
      code: 'Quote',
      shiftKey: '"',
      key: '\''
    },
    'AltGraph': {
      keyCode: 225,
      code: 'AltGraph',
      key: 'AltGraph'
    },
    'Props': {
      keyCode: 247,
      code: 'Props',
      key: 'CrSel'
    },
    'Cancel': {
      keyCode: 3,
      key: 'Cancel',
      code: 'Abort'
    },
    'Clear': {
      keyCode: 12,
      key: 'Clear',
      code: 'Numpad5',
      location: 3
    },
    'Shift': {
      keyCode: 16,
      key: 'Shift',
      code: 'ShiftLeft',
      location: 1
    },
    'Control': {
      keyCode: 17,
      key: 'Control',
      code: 'ControlLeft',
      location: 1
    },
    'Alt': {
      keyCode: 18,
      key: 'Alt',
      code: 'AltLeft',
      location: 1
    },
    'Accept': {
      keyCode: 30,
      key: 'Accept'
    },
    'ModeChange': {
      keyCode: 31,
      key: 'ModeChange'
    },
    ' ': {
      keyCode: 32,
      key: ' ',
      code: 'Space'
    },
    'Print': {
      keyCode: 42,
      key: 'Print'
    },
    'Execute': {
      keyCode: 43,
      key: 'Execute',
      code: 'Open'
    },
    '\u0000': {
      keyCode: 46,
      key: '\u0000',
      code: 'NumpadDecimal',
      location: 3
    },
    'a': {
      keyCode: 65,
      key: 'a',
      code: 'KeyA'
    },
    'b': {
      keyCode: 66,
      key: 'b',
      code: 'KeyB'
    },
    'c': {
      keyCode: 67,
      key: 'c',
      code: 'KeyC'
    },
    'd': {
      keyCode: 68,
      key: 'd',
      code: 'KeyD'
    },
    'e': {
      keyCode: 69,
      key: 'e',
      code: 'KeyE'
    },
    'f': {
      keyCode: 70,
      key: 'f',
      code: 'KeyF'
    },
    'g': {
      keyCode: 71,
      key: 'g',
      code: 'KeyG'
    },
    'h': {
      keyCode: 72,
      key: 'h',
      code: 'KeyH'
    },
    'i': {
      keyCode: 73,
      key: 'i',
      code: 'KeyI'
    },
    'j': {
      keyCode: 74,
      key: 'j',
      code: 'KeyJ'
    },
    'k': {
      keyCode: 75,
      key: 'k',
      code: 'KeyK'
    },
    'l': {
      keyCode: 76,
      key: 'l',
      code: 'KeyL'
    },
    'm': {
      keyCode: 77,
      key: 'm',
      code: 'KeyM'
    },
    'n': {
      keyCode: 78,
      key: 'n',
      code: 'KeyN'
    },
    'o': {
      keyCode: 79,
      key: 'o',
      code: 'KeyO'
    },
    'p': {
      keyCode: 80,
      key: 'p',
      code: 'KeyP'
    },
    'q': {
      keyCode: 81,
      key: 'q',
      code: 'KeyQ'
    },
    'r': {
      keyCode: 82,
      key: 'r',
      code: 'KeyR'
    },
    's': {
      keyCode: 83,
      key: 's',
      code: 'KeyS'
    },
    't': {
      keyCode: 84,
      key: 't',
      code: 'KeyT'
    },
    'u': {
      keyCode: 85,
      key: 'u',
      code: 'KeyU'
    },
    'v': {
      keyCode: 86,
      key: 'v',
      code: 'KeyV'
    },
    'w': {
      keyCode: 87,
      key: 'w',
      code: 'KeyW'
    },
    'x': {
      keyCode: 88,
      key: 'x',
      code: 'KeyX'
    },
    'y': {
      keyCode: 89,
      key: 'y',
      code: 'KeyY'
    },
    'z': {
      keyCode: 90,
      key: 'z',
      code: 'KeyZ'
    },
    'Meta': {
      keyCode: 91,
      key: 'Meta',
      code: 'MetaLeft',
      location: 1
    },
    '*': {
      keyCode: 106,
      key: '*',
      code: 'NumpadMultiply',
      location: 3
    },
    '+': {
      keyCode: 107,
      key: '+',
      code: 'NumpadAdd',
      location: 3
    },
    '-': {
      keyCode: 109,
      key: '-',
      code: 'NumpadSubtract',
      location: 3
    },
    '/': {
      keyCode: 111,
      key: '/',
      code: 'NumpadDivide',
      location: 3
    },
    ';': {
      keyCode: 186,
      key: ';',
      code: 'Semicolon'
    },
    '=': {
      keyCode: 187,
      key: '=',
      code: 'Equal'
    },
    ',': {
      keyCode: 188,
      key: ',',
      code: 'Comma'
    },
    '.': {
      keyCode: 190,
      key: '.',
      code: 'Period'
    },
    '`': {
      keyCode: 192,
      key: '`',
      code: 'Backquote'
    },
    '[': {
      keyCode: 219,
      key: '[',
      code: 'BracketLeft'
    },
    '\\': {
      keyCode: 220,
      key: '\\',
      code: 'Backslash'
    },
    ']': {
      keyCode: 221,
      key: ']',
      code: 'BracketRight'
    },
    '\'': {
      keyCode: 222,
      key: '\'',
      code: 'Quote'
    },
    'Attn': {
      keyCode: 246,
      key: 'Attn'
    },
    'CrSel': {
      keyCode: 247,
      key: 'CrSel',
      code: 'Props'
    },
    'ExSel': {
      keyCode: 248,
      key: 'ExSel'
    },
    'EraseEof': {
      keyCode: 249,
      key: 'EraseEof'
    },
    'Play': {
      keyCode: 250,
      key: 'Play'
    },
    'ZoomOut': {
      keyCode: 251,
      key: 'ZoomOut'
    },
    ')': {
      keyCode: 48,
      key: ')',
      code: 'Digit0'
    },
    '!': {
      keyCode: 49,
      key: '!',
      code: 'Digit1'
    },
    '@': {
      keyCode: 50,
      key: '@',
      code: 'Digit2'
    },
    '#': {
      keyCode: 51,
      key: '#',
      code: 'Digit3'
    },
    '$': {
      keyCode: 52,
      key: '$',
      code: 'Digit4'
    },
    '%': {
      keyCode: 53,
      key: '%',
      code: 'Digit5'
    },
    '^': {
      keyCode: 54,
      key: '^',
      code: 'Digit6'
    },
    '&': {
      keyCode: 55,
      key: '&',
      code: 'Digit7'
    },
    '(': {
      keyCode: 57,
      key: '(',
      code: 'Digit9'
    },
    'A': {
      keyCode: 65,
      key: 'A',
      code: 'KeyA'
    },
    'B': {
      keyCode: 66,
      key: 'B',
      code: 'KeyB'
    },
    'C': {
      keyCode: 67,
      key: 'C',
      code: 'KeyC'
    },
    'D': {
      keyCode: 68,
      key: 'D',
      code: 'KeyD'
    },
    'E': {
      keyCode: 69,
      key: 'E',
      code: 'KeyE'
    },
    'F': {
      keyCode: 70,
      key: 'F',
      code: 'KeyF'
    },
    'G': {
      keyCode: 71,
      key: 'G',
      code: 'KeyG'
    },
    'H': {
      keyCode: 72,
      key: 'H',
      code: 'KeyH'
    },
    'I': {
      keyCode: 73,
      key: 'I',
      code: 'KeyI'
    },
    'J': {
      keyCode: 74,
      key: 'J',
      code: 'KeyJ'
    },
    'K': {
      keyCode: 75,
      key: 'K',
      code: 'KeyK'
    },
    'L': {
      keyCode: 76,
      key: 'L',
      code: 'KeyL'
    },
    'M': {
      keyCode: 77,
      key: 'M',
      code: 'KeyM'
    },
    'N': {
      keyCode: 78,
      key: 'N',
      code: 'KeyN'
    },
    'O': {
      keyCode: 79,
      key: 'O',
      code: 'KeyO'
    },
    'P': {
      keyCode: 80,
      key: 'P',
      code: 'KeyP'
    },
    'Q': {
      keyCode: 81,
      key: 'Q',
      code: 'KeyQ'
    },
    'R': {
      keyCode: 82,
      key: 'R',
      code: 'KeyR'
    },
    'S': {
      keyCode: 83,
      key: 'S',
      code: 'KeyS'
    },
    'T': {
      keyCode: 84,
      key: 'T',
      code: 'KeyT'
    },
    'U': {
      keyCode: 85,
      key: 'U',
      code: 'KeyU'
    },
    'V': {
      keyCode: 86,
      key: 'V',
      code: 'KeyV'
    },
    'W': {
      keyCode: 87,
      key: 'W',
      code: 'KeyW'
    },
    'X': {
      keyCode: 88,
      key: 'X',
      code: 'KeyX'
    },
    'Y': {
      keyCode: 89,
      key: 'Y',
      code: 'KeyY'
    },
    'Z': {
      keyCode: 90,
      key: 'Z',
      code: 'KeyZ'
    },
    ':': {
      keyCode: 186,
      key: ':',
      code: 'Semicolon'
    },
    '<': {
      keyCode: 188,
      key: '<',
      code: 'Comma'
    },
    '_': {
      keyCode: 189,
      key: '_',
      code: 'Minus'
    },
    '>': {
      keyCode: 190,
      key: '>',
      code: 'Period'
    },
    '?': {
      keyCode: 191,
      key: '?',
      code: 'Slash'
    },
    '~': {
      keyCode: 192,
      key: '~',
      code: 'Backquote'
    },
    '{': {
      keyCode: 219,
      key: '{',
      code: 'BracketLeft'
    },
    '|': {
      keyCode: 220,
      key: '|',
      code: 'Backslash'
    },
    '}': {
      keyCode: 221,
      key: '}',
      code: 'BracketRight'
    },
    '"': {
      keyCode: 222,
      key: '"',
      code: 'Quote'
    }
  };

  function getTextRange($element) {
    let win = $element.ownerDocument.defaultView;
    let sel = win.getSelection();
    let start, end;
    if ('selectionStart' in $element) {
      ({
        selectionStart: start,
        selectionEnd: end
      } = $element);
    } else if (sel.containsNode($element, true) || $element.contains(sel.anchorNode) || $element.contains(sel.focusNode)) {
      let range = sel.getRangeAt(0);
      let caretRange = range.cloneRange();
      caretRange.selectNodeContents($element);
      caretRange.setEnd(range.startContainer, range.startOffset);
      start = caretRange.toString().length;
      caretRange.setEnd(range.endContainer, range.endOffset);
      end = caretRange.toString().length;
    }
    if (start || end) return [start, end];
  }
  function exec$2($element, parsed, _ref) {
    let {
      range,
      replace
    } = _ref;
    let {
      text: char,
      event
    } = parsed;
    let cancelled = !dispatch($element, 'keydown', event);
    if (cancelled || !(char || event.key === 'Backspace' || event.key === 'Delete')) return;
    let isInput = /^(input|textarea)$/i.test($element.tagName);
    if (!isInput && !$element.isContentEditable) return;
    let descr = isInput && getOwnPropertyDescriptor($element, 'value');
    if (descr) delete $element.value;
    event = char ? assign({
      charCode: char.charCodeAt(0)
    }, event) : event;
    cancelled = !dispatch($element, 'keypress', event) || !dispatch($element, 'beforeinput', event);
    if (!cancelled) {
      let value = isInput ? $element.value : $element.textContent;
      range = replace ? [0, value.length] : range;
      range = [].concat(range || getTextRange($element) || value.length);
      if (range.length === 1) range[1] = range[0];
      if (event.key === 'Backspace') range[0] -= 1;
      if (event.key === 'Delete') range[1] += 1;
      value = value.slice(0, range[0]) + char + value.slice(range[1]);
      if (isInput) $element.value = value;else $element.textContent = value;
      dispatch($element, 'input', assign({
        cancelable: false
      }, event));
    }
    if (descr) defineProperty($element, 'value', descr);
  }
  function keydown(key) {
    let options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    return k.press(key, true, this.assert.exists().exec(function ($element) {
      exec$2($element, k.parse(this, key), options);
    }));
  }

  function exec$1($element, _ref) {
    let {
      event
    } = _ref;
    dispatch($element, 'keyup', {
      cancelable: false,
      ...event
    });
  }
  function keyup(key) {
    return k.press(key, false, this.assert.exists().exec(function ($element) {
      exec$1($element, k.parse(this, key));
    }));
  }

  async function exec($element, parsed, _ref) {
    let {
      range,
      replace,
      delay = 1
    } = _ref;
    range = range != null && [].concat(range);
    for (let i = 0, l = parsed.length; i < l; i++) {
      let ra = range && (i === 0 ? range : range[0] + i);
      let re = i === 0 && replace;
      exec$2($element, parsed[i], {
        range: ra,
        replace: re
      });
      exec$1($element, parsed[i]);
      if (delay && i !== l - 1) await new Promise(r => setTimeout(r, delay));
    }
  }
  function press(keys) {
    let options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    return this.assert.exists().exec(async function ($element) {
      let parsed = [].concat(keys).map(c => k.parse(this, c));
      await exec($element, parsed, options);
    });
  }

  function scroll() {
    let options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    let next = this;
    let {
      top = options.y,
      left = options.x,
      wheel = false,
      frequency = 1
    } = options;
    if (top == null && left == null) throw InteractorError('missing scroll direction');
    if (left != null) next = next.assert.scrollableX();
    if (top != null) next = next.assert.scrollableY();
    return next.exec($el => {
      for (let i = 1; i <= frequency; i++) {
        let cancelled = false;
        if (wheel) cancelled = !dispatch($el, 'wheel', {});
        if (!cancelled) {
          if (top) $el.scrollTop = top / frequency * i;
          if (left) $el.scrollLeft = left / frequency * i;
          dispatch($el, 'scroll', {
            cancelable: false
          });
        }
      }
    });
  }

  function options($el, selector, each) {
    let multi = Array.isArray(selector);
    if (!('options' in $el)) throw InteractorError('%{@} is not a select element');else if (multi && !$el.multiple) throw InteractorError('%{@} is not a multi select element');
    if ($el.multiple) {
      if (multi) selector.map(s => each(this.$(s), s));else this.$$(selector).map($o => each($o, selector));
    } else {
      each(this.$(selector), selector);
    }
  }
  function select(selector) {
    return this.assert.not.disabled().assert(function ($el) {
      options.call(this, $el, selector, ($opt, sel) => {
        if ($opt.disabled) throw InteractorError(`%{@ ${sel}} is disabled`);
      });
    }).exec(function ($el) {
      options.call(this, $el, selector, $opt => {
        $opt.selected = $el.multiple ? !$opt.selected : true;
        dispatch($el, 'input', {
          cancelable: false
        });
        dispatch($el, 'change', {
          cancelable: false
        });
      });
    });
  }

  function trigger(event) {
    let options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    return this.assert.exists().exec($el => dispatch($el, event, options));
  }

  function get$b() {
    return this.$() === document.activeElement;
  }

  var focused = /*#__PURE__*/Object.freeze({
    __proto__: null,
    get: get$b
  });

  function type(text) {
    let {
      range,
      replace,
      delay,
      focus = true,
      blur = true,
      change = blur
    } = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    return this.assert.focusable().exec(async function ($element) {
      if (focus && !get$b.call(this)) exec$3($element);
      let parsed = [...text].map(c => k.parse(this, c));
      await exec($element, parsed, {
        range,
        replace,
        delay
      });
      if (change && /^(input|textarea)$/i.test($element.tagName)) dispatch($element, 'change', {
        cancelable: false
      });
      if (blur && get$b.call(this)) $element.blur();
    });
  }

  function isCheckboxNotRadio($el) {
    if ($el.type === 'radio') throw InteractorError('%{@} is a radio button which cannot be unchecked');else if ($el.type !== 'checkbox') throw InteractorError('%{@} is not a checkbox');
  }
  function uncheck() {
    return this.assert(isCheckboxNotRadio).assert.not.disabled().assert.checked().exec($el => $el.click());
  }

  var actions = /*#__PURE__*/Object.freeze({
    __proto__: null,
    blur: blur,
    check: check,
    click: click,
    focus: focus,
    keydown: keydown,
    keyup: keyup,
    press: press,
    scroll: scroll,
    select: select,
    trigger: trigger,
    type: type,
    uncheck: uncheck
  });

  function value$4(attr) {
    return this.$().getAttribute(attr);
  }

  var attribute = /*#__PURE__*/Object.freeze({
    __proto__: null,
    value: value$4
  });

  function get$a() {
    return !!this.$().checked;
  }

  var checked = /*#__PURE__*/Object.freeze({
    __proto__: null,
    get: get$a
  });

  function newEmptyInteractor(inst) {
    let I, proto;
    while ((proto = getPrototypeOf(proto || inst)) !== Object.prototype) I = proto.constructor;
    return m.new(I(), 'selector', null);
  }
  function value$3() {
    let selector = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : '';
    let parent = selector && this;
    if (!parent) {
      ({
        parent,
        selector
      } = m.get(this));
      parent = parent || newEmptyInteractor(this);
    }
    return parent.$$(selector).length;
  }
  const assert$6 = assertion(function (selector) {
    let expected = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : selector;
    selector = expected !== selector ? selector : '';
    let actual = value$3.call(this, selector);
    return {
      message: `found ${actual} element${actual === 1 ? '' : 's'} ` + `matching %{@ ${selector}} but expected %{- ${expected}|not to}`,
      result: actual === expected
    };
  });

  var count = /*#__PURE__*/Object.freeze({
    __proto__: null,
    assert: assert$6,
    value: value$3
  });

  function get$9() {
    return !!this.$().disabled;
  }

  var disabled = /*#__PURE__*/Object.freeze({
    __proto__: null,
    get: get$9
  });

  function get$8() {
    try {
      return !!this.$();
    } catch (e) {
      return false;
    }
  }
  const assert$5 = assertion(get$8, result => ({
    message: '%{@} %{- does not exist|exists}',
    result
  }));

  var exists = /*#__PURE__*/Object.freeze({
    __proto__: null,
    assert: assert$5,
    get: get$8
  });

  function get$7() {
    return !this.disabled && !!(~this.$().tabIndex || this.$().isContentEditable);
  }
  function assert$4(expected) {
    let result = get$7.call(this);
    if (expected) this.assert.not.disabled();
    if (result !== expected) throw InteractorError('%{@} is %{- not} focusable');
  }

  var focusable = /*#__PURE__*/Object.freeze({
    __proto__: null,
    assert: assert$4,
    get: get$7
  });

  function value$2(selector) {
    return this.$().matches(selector);
  }
  const assert$3 = assertion(value$2, (result, selector) => ({
    message: `%{@} %{- does not match|matches} ${selector} ` + '%{- |but expected it not to}',
    result
  }));

  var matches = /*#__PURE__*/Object.freeze({
    __proto__: null,
    assert: assert$3,
    value: value$2
  });

  const LAYOUT_PROPS = ['scrollTop', 'scrollLeft', 'scrollWidth', 'scrollHeight', 'clientTop', 'clientLeft', 'clientWidth', 'clientHeight'];
  function value$1(prop) {
    if (LAYOUT_PROPS.includes(prop)) hasLayoutEngine(this, 'Layout');
    return this.$()[prop];
  }

  var property = /*#__PURE__*/Object.freeze({
    __proto__: null,
    value: value$1
  });

  function isScrollableX(inst, $el) {
    return hasLayoutEngine(inst, 'Overflow') && $el.scrollWidth > $el.clientWidth;
  }
  function get$6() {
    return isScrollableX(this, this.$());
  }
  const assert$2 = assertion(get$6, result => ({
    message: '%{@} has %{- no} overflow-x',
    result
  }));

  var scrollableX = /*#__PURE__*/Object.freeze({
    __proto__: null,
    assert: assert$2,
    get: get$6,
    isScrollableX: isScrollableX
  });

  function isScrollableY(inst, $el) {
    return hasLayoutEngine(inst, 'Overflow') && $el.scrollHeight > $el.clientHeight;
  }
  function get$5() {
    return isScrollableY(this, this.$());
  }
  const assert$1 = assertion(get$5, result => ({
    message: '%{@} has %{- no} overflow-y',
    result
  }));

  var scrollableY = /*#__PURE__*/Object.freeze({
    __proto__: null,
    assert: assert$1,
    get: get$5,
    isScrollableY: isScrollableY
  });

  function get$4() {
    let $el = this.$();
    return isScrollableY(this, $el) || isScrollableX(this, $el);
  }
  const assert = assertion(get$4, result => ({
    message: '%{@} has %{- no} overflow',
    result
  }));

  var scrollable = /*#__PURE__*/Object.freeze({
    __proto__: null,
    assert: assert,
    get: get$4
  });

  function get$3() {
    return !!this.$().selected;
  }

  var selected = /*#__PURE__*/Object.freeze({
    __proto__: null,
    get: get$3
  });

  function get$2() {
    return hasLayoutEngine(this, 'Text content') ? this.$().innerText : this.$().textContent;
  }

  var text$1 = /*#__PURE__*/Object.freeze({
    __proto__: null,
    get: get$2
  });

  function get$1() {
    return this.$().value;
  }

  var value = /*#__PURE__*/Object.freeze({
    __proto__: null,
    get: get$1
  });

  function isHiddenByStyles(element) {
    let style = getComputedStyle(element);
    return style.display === 'none' || style.visibility !== 'visible' || style.opacity < 0.1;
  }
  function isInsideViewport(element) {
    let {
      top,
      left,
      bottom,
      right,
      width,
      height
    } = element.getBoundingClientRect();
    if (width === 0 || height === 0 || bottom < 0 || right < 0) return false;
    let {
      innerHeight,
      innerWidth
    } = element.ownerDocument.defaultView;
    return top < innerHeight && left < innerWidth;
  }
  function isElementAtPoint(element, x, y) {
    let pointer = element.ownerDocument.elementFromPoint(x, y);
    if (pointer === element) return true;
    while (pointer = pointer?.parentNode) if (pointer === element) return true;
    return false;
  }
  function isElementOnTop(element) {
    let {
      top,
      left,
      bottom,
      right,
      width,
      height
    } = element.getBoundingClientRect();
    let [cx, cy] = [left + width / 2, top + height / 2];
    return isElementAtPoint(element, cx, cy) || isElementAtPoint(element, cx, top + 1) || isElementAtPoint(element, cx, bottom - 1) || isElementAtPoint(element, left + 1, cy) || isElementAtPoint(element, right - 1, cy);
  }
  function get() {
    let element = this.$();
    hasLayoutEngine(this, 'Visibility');
    if (isHiddenByStyles(element)) return false;
    if (!isInsideViewport(element)) return false;
    if (!isElementOnTop(element)) return false;
    return true;
  }

  var visible = /*#__PURE__*/Object.freeze({
    __proto__: null,
    get: get
  });

  var properties = /*#__PURE__*/Object.freeze({
    __proto__: null,
    attribute: attribute,
    checked: checked,
    count: count,
    disabled: disabled,
    exists: exists,
    focusable: focusable,
    focused: focused,
    matches: matches,
    property: property,
    scrollable: scrollable,
    scrollableX: scrollableX,
    scrollableY: scrollableY,
    selected: selected,
    text: text$1,
    value: value,
    visible: visible
  });

  function xpath(path) {
    return assign(($node, multiple) => {
      let result = document.evaluate(path, $node, null, multiple ? 5 : 9, null);
      let arr, n;
      if (multiple) {
        arr = [];
        while (n = result.iterateNext()) arr.push(n);
        return arr;
      }
      return result.singleNodeValue;
    }, {
      toString: () => `xpath(${path})`
    });
  }
  function text(string) {
    let selector = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : '*';
    return assign(($node, multiple) => {
      let result;
      for (let $el of $node.querySelectorAll(selector)) {
        let text = $el.innerText ?? $el.textContent;
        if (!string || text === string) {
          if (multiple) result = (result || []).concat($el);else return $el;
        }
      }
      return result;
    }, {
      toString: () => `"${string}"`
    });
  }
  function nth(n, selector) {
    let of = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 'child';
    let isInt = Number.isInteger(n);
    let last = isInt && n < 0;
    if (last) {
      of = `last-${of}`;
      n = Math.abs(n);
    }
    return assign(($node, multiple) => {
      let sel = n != null ? `${selector}:nth-${of}(${n})` : selector;
      return multiple ? $node.querySelectorAll(sel) : $node.querySelector(sel);
    }, {
      toString: () => {
        if (n == null) return selector;
        if (!isInt) return `the nth(${n}) ${selector}`;
        if (last && n === 1) return `the last ${selector}`;
        let teen = Array.from(`${n}`).reverse()[1] === '1';
        let th = !teen && ['', 'st', 'nd', 'rd'][n % 10] || 'th';
        return `the ${n}${th} ${last ? 'last' : ''} ${selector}`;
      }
    });
  }

  var selectors = /*#__PURE__*/Object.freeze({
    __proto__: null,
    nth: nth,
    text: text,
    xpath: xpath
  });

  function Interactor(selector, properties) {
    if (properties) return Interactor.extend(properties)(selector);else if (!(this instanceof Interactor)) return new Interactor(selector);
    m.set(this, {
      selector: this.constructor.selector(selector),
      timeout: this.constructor.timeout,
      interval: 10,
      keyboard: {},
      queue: [],
      nested: false,
      parent: null,
      top: true
    });
    defineProperty(this, 'assert', {
      configurable: true,
      get: () => {
        let a = InteractorAssert(this);
        defineProperty(this, 'assert', {
          configurable: true,
          value: a
        });
        return a;
      }
    });
  }
  defineProperties(Interactor, {
    find: {
      get: function () {
        return assign(m.set(selector => {
          return m.new(this(selector), 'nested', true);
        }, 'constructor', this), selectors);
      }
    },
    extend: {
      value: extend
    },
    assertion: {
      value: assertion
    },
    Error: {
      value: InteractorError
    },
    name: {
      writable: true,
      value: ''
    },
    timeout: {
      writable: true,
      value: 2000
    },
    selector: {
      configurable: true,
      get: () => s => s,
      set: function setSelectorFn(selector) {
        defineProperty(this, 'selector', {
          configurable: true,
          set: setSelectorFn,
          get: () => s => {
            let sel = typeof selector === 'function' ? selector(s) : s ?? selector;
            let toString = sel && hasOwn(sel, 'toString') ? sel.toString : () => s ?? (!this.name && sel?.toString() || '');
            return assign(typeof sel === 'function' ? sel : () => sel, {
              toString
            });
          }
        });
      }
    },
    dom: {
      configurable: true,
      get: () => window,
      set: function setDOM(win) {
        defineProperty(this, 'dom', {
          configurable: true,
          get: typeof win === 'function' ? win : () => win,
          set: setDOM
        });
      }
    },
    suppressLayoutEngineWarning: {
      configurable: true,
      get: function () {
        return !!getPrototypeOf(this).constructor.suppressLayoutEngineWarning;
      },
      set: function suppressLayoutEngineWarning(bool) {
        defineProperty(this, 'suppressLayoutEngineWarning', {
          configurable: true,
          get: () => !!bool,
          set: suppressLayoutEngineWarning
        });
      }
    }
  });
  assign(Interactor.prototype, {
    $(selector) {
      return query.call(this, selector);
    },
    $$(selector) {
      return query.call(this, selector, true);
    },
    timeout(ms) {
      return ms ? m.new(m.top(this), 'timeout', ms) : m.get(m.top(this), 'timeout');
    },
    find(selector) {
      let q = m.get(selector, 'queue');
      let i = q ? selector : Interactor(selector);
      if ((q || m.get(i, 'queue')).length) {
        throw InteractorError('the provided interactor must not have queued actions');
      }
      return m.set(m.new(i, 'parent', this), 'nested', true);
    },
    assert(assertion) {
      return m.new(m.top(this), 'queue', {
        type: 'assert',
        fn: assertion,
        ctx: this
      });
    },
    exec(callback) {
      if (m.get(callback, 'queue')) {
        let {
          queue,
          keyboard
        } = m.get(callback);
        queue = queue.map(a => assign({}, a, {
          ctx: m.new(a.ctx, 'parent', this)
        }));
        return m.new(m.top(this), {
          keyboard,
          queue
        });
      } else {
        return m.new(m.top(this), 'queue', {
          type: 'exec',
          fn: callback,
          ctx: this
        });
      }
    },
    catch(handler) {
      if (typeof handler === 'string') {
        let message = handler;
        handler = err => {
          if (err.name === 'InteractorError') throw err.format(message);else throw err;
        };
      }
      return m.new(m.top(this), 'queue', {
        type: 'catch',
        fn: handler,
        ctx: this
      });
    },
    then() {
      let {
        queue,
        timeout,
        interval
      } = m.get(this);
      let assertion;
      return queue.reduce((promise, action, i) => {
        let {
          type,
          ctx,
          fn
        } = action;
        let wrapper = fn => fn();
        if (!fn) type = null;
        let bind$ = (fn, f) => named(fn.name, () => wrapper(() => {
          return f?.(), fn.length ? fn.call(ctx, ctx.$()) : fn.call(ctx);
        }));
        let onError = err => {
          throw err.name === 'InteractorError' ? err.bind(ctx) : err;
        };
        if (type === 'assert') assertion = bind$(fn, assertion);
        if (assertion && (type !== 'assert' || i === queue.length - 1)) {
          let converge = when(assertion, {
            remains: action.remains,
            interval,
            timeout
          });
          promise = promise.then(converge).catch(onError);
          assertion = null;
        }
        if (type === 'exec') {
          wrapper = this.constructor.executionMiddleware ?? wrapper;
          promise = promise.then(bind$(fn)).catch(onError);
        } else if (type === 'catch') {
          promise = promise.catch(fn.bind(ctx));
        }
        return promise;
      }, Promise.resolve()).then(...arguments);
    },
    toString() {
      let name = this.constructor.name;
      let {
        parent,
        nested,
        selector
      } = m.get(this);
      let string = selector?.toString() || '';
      if (!string) string = name || '';else if (name) string += ` ${name}`;
      if (nested && m.get(parent, 'selector')) string += `${string && ' within '}${parent}`;
      return string;
    }
  });
  defineInteractorProperties(Interactor.prototype, assign({}, actions, properties));

  assign(Interactor, map(actions, (action, name) => {
    return named(name, function (selector) {
      for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
        args[_key - 1] = arguments[_key];
      }
      return action.apply(m.get(selector, 'queue') ? selector : Interactor.find(selector), args);
    });
  }));
  assign(Interactor, map(properties, (_ref, name) => {
    let {
      get,
      value,
      assert
    } = _ref;
    return named(name, function () {
      for (var _len2 = arguments.length, args = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
        args[_key2] = arguments[_key2];
      }
      let selector = get || value && args.length > value.length ? args.shift() : '';
      let assertion = assert || createAssert(name, get || value);
      let ctx = i => selector ? i.find(selector) : i;
      let thennable = when(() => {
        if (!selector) throw InteractorError('an element selector is required when awaiting on properties');
        return (get || value).apply(Interactor.find(selector), args);
      });
      return {
        then: thennable.then,
        catch: thennable.catch,
        get() {
          return (get || value).apply(ctx(this), args);
        },
        assert(expected) {
          for (var _len3 = arguments.length, a = new Array(_len3 > 1 ? _len3 - 1 : 0), _key3 = 1; _key3 < _len3; _key3++) {
            a[_key3 - 1] = arguments[_key3];
          }
          return ctx(this).assert(named(name, function () {
            assertion.apply(this, [expected].concat(args, a));
          }));
        }
      };
    });
  }));

  exports.InteractorError = InteractorError;
  exports.assertion = assertion;
  exports.default = Interactor;
  exports.when = when;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
